;============================================================================
; MD64v20 - MONITOR for DRAGON64 version v2.0 - 2015-11-07
;============================================================================
; 6809 Debug monitor for use with NOICE09 for Dragon64 with 6809 only
;  If you need 6309 support, see moncoco.asm
;
; Copyright (c) 1992-1998 by John Hartman
;
; Modification History:
; 	27-Mar-2006 Gerard Noteboom for Coco 6809
;	07-Nov-2015 Pere Serrat - adapted for Dragon64 in RAM mode ($FA00 on)
;============================================================================
;
; To customize for a given target, you must change:
; - code in the hardware equates
; - the string  TSTG
; - the routine RESET
;
; You may or may not need to change 
; - routine GETCHAR
; - routine PUTCHAR, depending on how peculiar your UART is.
;
; This file has been assembled with Ciaran Anscomb's ASM6809
;
; To add mapped memory support (NOT USED):
; - 1) Define map port MAPREG here
; - 2) Define or import map port RAM image MAPIMG here if MAPREG is
;         write only.  (The application code must update MAPIMG before
;         outputing to MAPREG)
; - 3) Search for and modify MAPREG, MAPIMG, and REG_PAGE usage below
; - 4) In TSTG below edit "LOW AND HIGH LIMIT OF MAPPED MEM"
;         to appropriate range (typically 4000H to 07FFFH for two-bit MMU)
;
;============================================================================
; HARDWARE PLATFORM CUSTOMIZATIONS : Dragon 64
; Monitor is rewritten in Postion Independant Code.
; This will work only in MAP1 (all RAM) by using $FA00 onwards as ROM and RAM
; Top of RAM $FFFF (to be used just untill $FEFF)
; Basic ROM must be copied into RAM at $8000 and ends at $BFFF
; Disk Operating System should be copied at $C000
; Dragon Predefined Soft Interrupt Vector Area
;============================================================================
soft_vect   equ   $100     					; start of dragon vectors 
ramvct  		equ   soft_vect        			; redirection table in ram. 6 hooks
jump      	equ   $7e							; opcode for jmp
zero16b		equ	$8a							; zero 18 bits wide
combuf_size equ   128             			; data size for comm buffer
nvec        equ   8  		         		; number of vectors
getkey		equ	$8006							; rom routine
bascom		equ	$8371							; basic command interpreter

;============================================================================
; put your acia equates here  : synertek sy6551 in the dragon64
;============================================================================
r6551       equ   $ff04           			; acia base
ser_rxdata  equ   r6551+0
ser_txdata  equ   ser_rxdata
ser_status  equ   ser_rxdata+1
rxrdy       equ   $08
txrdy       equ	$10
; condition code bits
c       		equ   1
i       		equ   $10
f       		equ   $40
e       		equ   $80
; acia defines
rxdata   	equ   0         					; read  only
txdata   	equ   rxdata    					; write only
statr    	equ   1         					; read  only
rsetr    	equ   statr     					; write only
cmdr     	equ   2         					; r+w
ctrlr    	equ   3         					; r+w
;============================================================================
; hardware platform independent equates and code
; communications function codes.
;============================================================================
fn_get_stat equ   $ff    						; reply with device info
fn_read_mem equ   $fe    						; reply with data
fn_write_m  equ   $fd    						; reply with status (+/-)
fn_read_rg  equ   $fc    						; reply with registers
fn_write_rg equ   $fb    						; reply with status
fn_run_targ equ   $fa    						; reply (delayed) with registers
fn_set_byte equ   $f9    						; reply with data (truncate if error)
fn_in       equ   $f8    						; input from port
fn_out      equ   $f7    						; output to port
fn_min      equ   $f7    						; minimum recognized function code
fn_error    equ   $f0    						; error reply to unknown op-code

;===========================================================================
; monitor program
;===========================================================================
        		org   $fa00    					; start of monitor code
				put	$fa00
entry   												;  power on reset
; set cpu mode to safe state
reset	  		orcc  #i+f            			; interrupts off
        		sts	oldstack						; save dragon stack pointer
        		lds   #monstack       			; clean stack is happy stack
; acia initialization
iniacia  	ldx 	#r6551						; x= $ff04
         	clr   rsetr,x    					; $ff05 = $00 - reset, do not care data
				ldd	#$0b1f						; a=#$0b (no parity, no irq, enable rx, tx, /rts, /dtr)
														; b=#$1f (19200 baud, 8 bits , 1 stopbit)
         	std   cmdr,x     					; write to command reg + control reg

;===========================================================================
; initializations
;===========================================================================
; initialize ram interrupt vectors
        		leau  int_entry,pcr   			; address of default handler (pcr !)
        		ldx   #ramvec         			; pointer to ram vectors
        		ldb   #nvec           			; number of vectors
res10   		stu   ,x++            			; set vector
        		decb
        		bne   res10
; save dragon interrupt vectors to table
drgvct   	ldb   #18                 		; number of bytes
          	ldx   #ramvct						; pointer to dragon ram vectors
          	leay  oldvec,pcr          		; (pcr !)
				lbsr	cpyx2y						; copy b bytes from x to y
; dragon ram interrupt vectors redirection
softvct   	ldy   #6          				; build int.vect. table in ram
				ldu   #ramvct
				leax  vectbl,pcr  				; (pcr !)
jmprpt    	lda   #jump
				sta   ,u+
				tfr    x,d
				addd  ,x++
				std   ,u++
				leay  -1,y
				bne   jmprpt
; initialize user registers
        		ldd   #initstack
				ldx   #reg_u						; point to u register (low byte)
        		sta   -1,x              		; init users stack pointer msb (reg_sp)
        		stb   -2,x                 	; lsb
        		ldd   <zero16b						; get 16 bits zero
        		std   -4,x		             	; initial state (reg_state) is "reset" 
        												;    and initial page (reg_page) is set to 0
				ldb   #12							; number of bytes to set to zero
iniregs		sta   ,x+							; affects: reg_u(2), reg_y(2), reg_x(2), reg_b
				decb  								; reg_a, reg_dp, reg_cc, reg_pc(2)
				bne	iniregs						; not yet done, loopback
; initialize memory paging variables and hardware (if any)
        		lda   #e+i+f                  ; state "all regs pushed", no ints
        		sta   -3,x							; into reg_cc
; set function code for "go".  then if we reset after being told to go,
; we will come back with registers so user can see the crash
        		lda   #fn_run_targ
        		sta   ,x								; into combuf (first byte)

;===========================================================================
; read registers:  fn, len=0
; entry with a=function code, b=data size, x=combuf+2
;===========================================================================
read_regs
; enter here from swi after "run" and "step" to return task registers
return_regs	ldy   #task_regs              ; pointer to registers
        		ldb   #task_reg_sz            ; number of bytes
        		ldx   #combuf+1               ; pointer to return buffer
        		stb   ,x+                     ; save return data length
; copy the registers
				exg	x,y
				lbsr	cpyx2y						; copy b bytes from x to y
				exg	x,y
; compute checksum on buffer, and send to master, then return

;===========================================================================
; append checksum to combuf and send to master
;===========================================================================
send    		lbsr  checksum                ; get a=checksum, x->checksum location
        		nega
        		sta   0,x                     ; store negative of checksum
; send buffer to master
        		ldx   #combuf                 ; pointer to data
        		ldb   1,x                     ; length of data
        		addb  #3                      ; plus function, length, checksum
snd10   		lda   ,x+
        		lbsr   putchar                ; send a byte
        		decb
        		bne   snd10

;===========================================================================
; main loop  wait for command frame from master
; uses 6 bytes of stack including return address
;============================================================================
main    		lds   #monstack               ; clean stack is happy stack
				jsr	getkey						; get key pressed
				cmpa	#$03							; is it break?
				lbeq	exitmon						; yes, exit monitor
        		ldx   #combuf                 ; build message here
; first byte is a function code
        		lbsr  getchar                 ; get a function (6 bytes of stack)
        		bcs   main                    ; jif timeout: resync
        		cmpa  #fn_min
        		blo   main                    ; jif below min: illegal function
        		sta   ,x+                     ; save function code
; second byte is data byte count (may be zero)
        		lbsr  getchar                 ; get a length byte
        		bcs   main                    ; jif timeout: resync
        		cmpa  #combuf_size
        		bhi   main                    ; jif too long: illegal length
        		sta   ,x+                     ; save length
        		beq   ma80                    ; skip data loop if length = 0
; loop for data
        		tfr   a,b                     ; save length for loop
ma10    		lbsr  getchar                 ; get a data byte
        		bcs   main                    ; jif timeout: resync
        		sta   ,x+                     ; save data byte
        		decb
        		bne   ma10
; get the checksum
ma80    		lbsr  getchar                 ; get the checksum
        		bcs   main                    ; jif timeout: resync
        		pshs  a                       ; save checksum
; compare received checksum to that calculated on received buffer (sum should be 0)
        		lbsr  checksum
        		adda  ,s+                   	; add saved checksum to computed
        		bne   main                  	; jif bad checksum
; process the message.
        		ldx   #combuf
				ldd	,x++							; get function and numbytes
				ldu	#out_f7						; point to beginning of table	
				suba	#$f7							; subtract lowest accepted function number
				bcs	badfunc						; if a was lower than $f7
				asla									; each entry is a word
				jmp	[a,u]							; goto requested function
; error: unknown function.  complain
badfunc	  	lda   #fn_error
			  	sta   combuf          			; set function as "error"
			  	lda   #1								; mark status error (z=0)

;===========================================================================
; build status return with value from "a"
;===========================================================================
send_status	sta   combuf+2                ; set status
        		lda   #1
        		sta   combuf+1                ; set length
        		bra	send

;===========================================================================
; target status:  fn, len
; entry with a=function code, b=data size, x=combuf+2
;============================================================================
target_stat	leax  tstg,pcr                ; data for reply
        		ldy   #combuf+1               ; pointer to return buffer
        		ldb   #tstg_size              ; length of reply
        		stb   ,y+                     ; set size in reply buffer
				bsr	cpyx2y						; copy b bytes from x to y
; compute checksum on buffer, and send to master, then return
        		lbra  send

;===========================================================================
; copy b bytes from x to y
;===========================================================================
cpyx2y  		lda   ,x+                     ; get a byte from origin
        		sta   ,y+							; put at destination
        		decb									; decrement counter
        		bne   cpyx2y						; not finished, loopback
				rts									; return

;===========================================================================
; read memory:  fn, len, page, alo, ahi, nbytes
; entry with a=function code, b=data size, x=combuf+2
;===========================================================================
; get address
read_mem		lda   2,x                     ; msb of address in a
        		ldb   1,x                     ; lsb of address in b
        		tfr   d,y                     ; address in y
; prepare return buffer: fn (unchanged), len, data
        		ldb   3,x                     ; number of bytes to return
        		stb   combuf+1                ; return length = requested data
        		beq   glp90                   ; jif no bytes to get
; read the requested bytes from local memory
				exg	x,y
				bsr	cpyx2y						; copy b bytes from x to y
				exg	x,y
; compute checksum on buffer, and send to master, then return
glp90   		lbra  send

;===========================================================================
; write memory:  fn, len, page, alo, ahi, (len-3 bytes of data)
; entry with a=function code, b=data size, x=combuf+2
; uses 6 bytes of stack
;===========================================================================
write_mem	leax  1,x							; skip map	
; get address
        		ldb   ,x+                     ; lsb of address in b
        		lda   ,x+                     ; msb of address in a
        		tfr   d,y                     ; address in y
; compute number of bytes to write
        		ldb   combuf+1                ; number of bytes to return
        		subb  #3                      ; minus page and address
        		beq   wlp50                   ; jif no bytes to put
; write the specified bytes to local memory
        		pshs  b,x,y
				bsr	cpyx2y						; copy b bytes from x to y
; compare to see if the write worked
        		puls  b,x,y
wlp20   		lda   ,x+                     ; get byte just written
        		cmpa  ,y+
        		bne   wlp80                   ; br if write failed
        		decb
        		bne   wlp20
; write succeeded:  return status = 0
wlp50   		clra		                     ; return status = 0
        		bra   wlp90
; write failed:  return status = 1
wlp80   		lda   #1
; return ok status
wlp90   		bra  	send_status

;===========================================================================
; write registers:  fn, len, (register image)
; entry with a=function code, b=data size, x=combuf+2
;===========================================================================
write_regs	tstb                          ; number of bytes
        		beq   wrr80                   ; jif no registers
; copy the registers
        		ldy   #task_regs              ; pointer to registers
				bsr	cpyx2y						; copy b bytes from x to y
; return ok status
wrr80   		clra
        		bra  	send_status

;===========================================================================
; run target:  fn, len
; entry with a=function code, b=data size, x=combuf+2
;===========================================================================
; switch to user stack
run_target	ldx	#reg_x 			  			; point to the middle of the registers table         	
          	lda   -5,x                 	; back to user stack (reg_sp(2))
        		ldb   -6,x				   
        		tfr   d,s                     ; to s
; restore registers					  
        		lda   7,x		               ; msb user pc for rti (reg_pc(2))
        		ldb   6,x							; lsb
        		pshs  d					  
        		lda   -3,x				  			; reg_u(2) msb
        		ldb   -4,x				  			; lsb
        		pshs  d					  
        		lda   -1,x				  			; reg_y (2) msb
        		ldb   -2,x				  			; lsb
        		pshs  d					  
        		lda   1,x				  			; reg_x (2) msb
        		ldb   0,x				  			; lsb
        		pshs  d					  
				lda	2,x				  			; reg_b
        		ldb   4,x				  			; reg_dp
				pshs	d					  
        		lda   5,x                    	; save user condition codes for rti. get reg_cc
        		ora   #e                      ; _must_ be "all regs pushed"
				ldb	3,x							; get reg_a				  
        		pshs  d					  
; return to user
        		rti

;===========================================================================
; set target byte(s):  fn, len { (page, alow, ahigh, data), (...)... }
; entry with a=function code, b=data size, x=combuf+2
; return has fn, len, (data from memory locations)
; if error in insert (memory not writable), abort to return short data
; this function is used primarily to set and clear breakpoints
; uses 1 byte of stack
;===========================================================================
set_bytes	ldu   #combuf+1               ; pointer to return buffer
        		clra
        		sta   ,u+                     ; set return count as zero
        		lsrb
        		lsrb                          ; len/4 = number of bytes to set
        		beq   sb99                    ; jif no bytes (combuf+1 = 0)
; loop on inserting bytes
sb10    		pshs  b                       ; save loop counter
; get address
        		lda   2,x                     ; msb of address in a
        		ldb   1,x                     ; lsb of address in b
        		tfr   d,y                     ; memory address in y
; read current data at byte location
        		lda   0,y
; insert new data at byte location
        		ldb   3,x                     ; get byte to store
        		stb   0,y                     ; write target memory
; verify write
        		cmpb  0,y                     ; read target memory
        		puls  b                       ; restore loop count, cc's intact
        		bne   sb90                    ; br if insert failed: abort
; save target byte in return buffer
        		sta   ,u+
        		inc   combuf+1                ; count one return byte
; loop for next byte
        		leax  4,x                     ; step to next byte specifier
        		cmpb  combuf+1
        		bne   sb10                    ; *loop for all bytes
; return buffer with data from byte locations
sb90
; compute checksum on buffer, and send to master, then return
sb99    		lbra  send

;===========================================================================
; input from port:  fn, len, portaddresslo, pahi (=0)
; while the 6809 has no input or output instructions, we retain these
; to allow write-without-verify
; entry with a=function code, b=data size, x=combuf+2
;===========================================================================
; get port address
in_port 		lda   1,x                     ; msb of address in a
        		ldb   0,x                     ; lsb of address in b
        		tfr   d,y                     ; memory address in y
; read the requested byte from local memory
        		lda   0,y
; return byte read as "status"
        		lbra  send_status

;===========================================================================
; output to port  fn, len, portaddresslo, pahi (=0), data
; entry with a=function code, b=data size, x=combuf+2
;===========================================================================
; get port address
out_port		lda   1,x                     ; msb of address in a
        		ldb   0,x                     ; lsb of address in b
        		tfr   d,y                     ; memory address in y
; get data
        		lda   2,x
; write value to port
        		sta   0,y
; do not read port to verify (some i/o devices don't like it)
; return status of ok
        		clra
				lbra	send_status
				
;===========================================================================
; compute checksum on combuf.  combuf+1 has length of data,
; also include function byte and length byte
; returns:
;      a = checksum
;      x = pointer to next byte in buffer (checksum location)
;      b is scratched
;===========================================================================
checksum		ldx   #combuf                 ; pointer to buffer
        		ldb   1,x                     ; length of message
        		addb  #2                      ; plus function, length
        		clra	                        ; init checksum to 0
chk10   		adda  ,x+
        		decb
        		bne   chk10                   ; loop for all
        		rts                           ; return with checksum in a

;===========================================================================
; get a character into a
; return a=char, cy=0 if data received
;        cy=1 if timeout (0.5 seconds)
; uses 6 bytes of stack including return address
;============================================================================
getchar		pshs  x
        		ldx   <zero16b         			; long timeout
gc10    		leax  -1,x							 
        		beq   gc90            			; exit if timeout
        		lda   ser_status      			; a = ($ff05) - read device status
        		anda  #rxrdy						; a &= $08	 
        		beq   gc10            			; not ready yet.
; data received:  return cy=0. 				; data in a
        		clra                  			; cy=0
        		lda   ser_rxdata      			; a = ($ff04) - read data
        		puls  x,pc								 
; timeout:  return cy=1						 
gc90    		orcc  #c              			; cy=1
        		puls  x,pc

;============================================================================
; output character in a
; uses 5 bytes of stack including return address
;============================================================================
putchar		pshs  a
pc10    		lda   ser_status     			; a = ($ff05) - check tx status
        		anda  #txrdy         			; a &= $10 - rx ready ?
        		beq   pc10								 
        		puls  a									 
        		sta   ser_txdata     			; $ff04 = a - transmit char.
        		rts

;===========================================================================
; interrupt handlers to catch unused interrupts and traps
; registers are stacked.  jump through ram vector using x, type in a
; this will affect only interrupt routines looking for register values!
; our default handler uses the code in "a" as the processor state to be
; passed back to the host.
;===========================================================================
res_ent     lda   #7
            ldx   ramvec+0
            jmp   0,x
swi3_ent    lda   #6
            ldx   ramvec+2
            jmp   0,x
swi2_ent    lda   #5
            ldx   ramvec+4
            jmp   0,x
; may have only pc and cc's pushed (unless we were waiting for an interrupt)
; push all registers here for common entry (else we can't use our ram vector)
firq_ent    sta   reg_a       				; save a reg
            puls  a           				; get cc's from stack
            bita  #e
            bne   firq9       				; br if all registers pushed already
            pshs  u,y,x,dp,b  				; else push them now
            ldb   reg_a
            pshs  b
            ora   #e          				; set as "all regs pushed"
firq9       pshs  a           				; replace cc's
            lda   #4
            ldx   ramvec+6
            jmp   0,x
irq_ent     lda   #3
            ldx   ramvec+8
            jmp   0,x
nmi_ent     lda   #2
            ldx   ramvec+12
            jmp   0,x
swi_ent     lda   #1

;============================================================================
; common handler for default interrupt handlers
; enter with a=interrupt code = processor state
; all registers stacked, pc=next instruction
;============================================================================
int_entry	ldu	#reg_x			
				sta   -8,u 		      			; save state (reg_state)
; save registers from stack to reg block for return to master
; host wants least significant bytes first, so flip as necessary
				puls	d
        		sta   5,u	         			; condition codes (reg_cc)
        		stb   3,u	         			; reg_a
				puls	d					
        		sta   2,u	         			; reg_b
        		stb   4,u	         			; reg_dp
        		puls  d
        		sta   1,u	         			; msb reg_x
        		stb   0,u	         			; lsb
        		puls  d
        		sta   -1,u	         			; msb reg_y
        		stb   -2,u           			; lsb
       		puls  d
        		sta   -3,u	         			; msb reg_u
        		stb   -4,u	         			; lsb
; if this is a breakpoint (state = 1), then back up pc to point at swi
        		puls  x               			; pc after interrupt
        		lda   -8,u							; get reg_state
        		cmpa  #1
        		bne   notbp           			; br if not a breakpoint
        		leax  -1,x            			; else back up to point at swi location
notbp   		tfr   x,d             			; transfer pc to d
        		sta   7,u 		      	 		; msb reg_pc
        		stb   6,u	         			; lsb

;===========================================================================
; common continue point for all monitor entrances
; sp = user stack
;===========================================================================
enter_mon	tfr   s,d             			; user stack pointer
        		sta   reg_sp+1        			; save user's stack pointer (msb)
        		stb   reg_sp          			; lsb
; change to our own stack
        		lds   #monstack       			; and use ours instead
; operating system variables
        		clra		              			; ... or zero if unmapped target
        		sta   reg_page        			; save user's page
; return registers to master
        		lbra  return_regs

;===========================================================================
; exit back to dragon interpreter
;===========================================================================
exitmon		ldb	#18                  	; number of bytes to copy
          	ldy   #ramvct						; pointer to dragon ram vectors
          	leax  oldvec,pcr          		; (pcr !)
				lbsr	cpyx2y						; copy b bytes from x to y
				lds	oldstack						; get old stack value
				andcc	#$af							; enable interrupts
				jmp	bascom						; back to the basic interpreter

;===========================================================================
; indirect table for function despatch
;===========================================================================
out_f7		fdb	out_port
				fdb	in_port
				fdb	set_bytes
				fdb	run_target
				fdb	write_regs
				fdb	read_regs
				fdb	write_mem
				fdb	read_mem
				fdb	target_stat

;============================================================================
; vectors via ram for coco-2 -- all references in rom have to be pc-relative 
;============================================================================
vectbl     	fdb  	swi3_ent-vectbl-0  		; put in ram-location $101 after jmp-code (7e)
           	fdb  	swi2_ent-vectbl-2  		;                     $104
           	fdb   swi_ent-vectbl-4  		;                     $107
           	fdb   nmi_ent-vectbl-6  		;                     $10a
           	fdb   irq_ent-vectbl-8  		;                     $10d
           	fdb  	firq_ent-vectbl-10 		;                     $110
oldvec		rzb	18								; to contain the 6 dragon vectors (3 bytes each)

;======================================================================
; response string for get target status request
; reply describes target:
;============================================================================
tstg   	 	fcb   5                       		; 2: processor type = 6809
       	 	fcb   combuf_size             		; 3: size of communications buffer
       	 	fcb   0                       		; 4: no tasking support
       	 	fdb   0,0                     		; 5-8: low and high limit of mapped mem (none)
       	 	fcb   b1-b0                   		; 9:  breakpoint instr length
b0     	 	swi                           		; 10: breakpoint instruction
b1     	 	fcn   /D64-6809 monitor V2.0/  		; description
tstg_size   equ   *-tstg          					; size of string

;============================================================================
; data area in high ram
;============================================================================
ram_start   		    							; start of monitor ram
; ram interrupt vectors (first in seg for easy addressing, else move to
; their own seg)
ramvec      rzb   2*nvec
; old dragon stack pointer to be restored on exit
oldstack		rzb	2
; initial user stack (size and location is user option)
            rzb   64
initstack
; monitor stack (calculated use is at most 7 bytes.  leave plenty of spare)
            rzb   16
monstack
; target registers:  order must match that in trghc11.c
task_regs
reg_state   rzb   1
reg_page    rzb   1
reg_sp      rzb   2
reg_u       rzb   2
reg_y       rzb   2
reg_x       rzb   2
reg_b       rzb   1           				; b before a, so d is least sig. first
reg_a       rzb   1
reg_dp      rzb   1
reg_cc      rzb   1
reg_pc      rzb   2
task_reg_sz equ   *-task_regs
; communications buffer
; (must be at least as long as task_reg_sz.  at least 19 bytes recommended.
; larger values may improve speed of noice memory move commands.)
combuf      rzb   2+combuf_size+1 			; buffer also has fn, len, and check
ram_end     equ   *               			; address of top+1 of ram

; ----------------------------------------------------------------------------
			if dw4 > 0
				if ((*+9)%256 > 0)
						rzb	 256-((*+9)%256)	; make it dw4 compatible (length multiple of 256 bytes)
				endif
		 	endif
;-----------------------------------------------------------------------------------------------------------
dw4_end		equ	*
        		end   entry
